﻿using System;
using System.Linq;
using DA = HeuristicLab.Services.Hive.DataAccess;
using DT = HeuristicLab.Services.Hive.DataTransfer;

namespace HeuristicLab.Services.Hive {
  public static class Converter {

    #region Task
    public static DT.Task ToDto(this DA.Task source) {
      if (source == null) return null;
      return new DT.Task {
        Id = source.TaskId,
        State = source.State.ToDto(),
        ExecutionTime = TimeSpan.FromMilliseconds(source.ExecutionTimeMs),
        LastHeartbeat = source.LastHeartbeat,
        ParentTaskId = source.ParentTaskId,
        Priority = source.Priority,
        CoresNeeded = source.CoresNeeded,
        MemoryNeeded = source.MemoryNeeded,
        IsParentTask = source.IsParentTask,
        FinishWhenChildJobsFinished = source.FinishWhenChildJobsFinished,
        Command = source.Command.ToDto(),
        JobId = source.JobId,
        IsPrivileged = source.IsPrivileged,
        PluginsNeededIds = source.RequiredPlugins.Select(x => x.PluginId).ToList(),
        StateLog = source.StateLogs.Select(x => x.ToDto()).OrderBy(x => x.DateTime).ToList(),
        LastTaskDataUpdate = source.JobData == null ? DateTime.MinValue : source.JobData.LastUpdate
      };
    }

    public static DA.Task ToEntity(this DT.Task source) {
      if (source == null) return null;
      var result = new DA.Task();
      source.CopyToEntity(result);
      return result;
    }

    public static void CopyToEntity(this DT.Task source, DA.Task target) {
      if ((source == null) || (target == null)) return;
      target.TaskId = source.Id;
      target.State = source.State.ToEntity();
      target.ExecutionTimeMs = source.ExecutionTime.TotalMilliseconds;
      target.LastHeartbeat = source.LastHeartbeat;
      target.ParentTaskId = source.ParentTaskId;
      target.Priority = source.Priority;
      target.CoresNeeded = source.CoresNeeded;
      target.MemoryNeeded = source.MemoryNeeded;
      target.IsParentTask = source.IsParentTask;
      target.FinishWhenChildJobsFinished = source.FinishWhenChildJobsFinished;
      target.Command = source.Command.ToEntity();
      target.JobId = source.JobId;
      target.IsPrivileged = source.IsPrivileged;
      var ids = target.RequiredPlugins.Select(x => x.PluginId);
      target.RequiredPlugins.AddRange(source.PluginsNeededIds
        .Where(x => !ids.Contains(x))
        .Select(x => new DA.RequiredPlugin {
          PluginId = x
        })
      );
      target.StateLogs.AddRange(source.StateLog
        .Where(x => x.Id == Guid.Empty)
        .Select(x => x.ToEntity())
      );
      // result.JobData missing
      // result.AssignedResources missing
    }
    #endregion

    #region TaskData
    public static DT.TaskData ToDto(this DA.TaskData source) {
      if (source == null) return null;
      return new DT.TaskData {
        TaskId = source.TaskId,
        Data = source.Data,
        LastUpdate = source.LastUpdate
      };
    }

    public static DA.TaskData ToEntity(this DT.TaskData source) {
      if (source == null) return null;
      var result = new DA.TaskData();
      source.CopyToEntity(result);
      return result;
    }

    public static void CopyToEntity(this DT.TaskData source, DA.TaskData target) {
      if ((source == null) || (target == null)) return;
      target.TaskId = source.TaskId;
      target.Data = source.Data;
      target.LastUpdate = source.LastUpdate;
    }
    #endregion

    #region Job
    public static DT.Job ToDto(this DA.Job source) {
      return new DT.Job {
        Id = source.JobId,
        Description = source.Description,
        Name = source.Name,
        OwnerUserId = source.OwnerUserId,
        DateCreated = source.DateCreated,
        ResourceNames = source.ResourceIds
      };
    }

    public static DA.Job ToEntity(this DT.Job source) {
      if (source == null) return null;
      var result = new DA.Job();
      source.CopyToEntity(result);
      return result;
    }

    public static void CopyToEntity(this DT.Job source, DA.Job target) {
      if ((source == null) || (target == null)) return;
      target.JobId = source.Id;
      target.Description = source.Description;
      target.Name = source.Name;
      target.OwnerUserId = source.OwnerUserId;
      target.DateCreated = source.DateCreated;
      target.ResourceIds = source.ResourceNames;
    }
    #endregion

    #region JobPermission
    public static DT.JobPermission ToDto(this DA.JobPermission source) {
      if (source == null) return null;
      return new DT.JobPermission {
        JobId = source.JobId,
        GrantedUserId = source.GrantedUserId,
        GrantedByUserId = source.GrantedByUserId,
        Permission = source.Permission.ToDto()
      };
    }

    public static DA.JobPermission ToEntity(this DT.JobPermission source) {
      if (source == null) return null;
      var result = new DA.JobPermission();
      source.CopyToEntity(result);
      return result;
    }

    public static void CopyToEntity(this DT.JobPermission source, DA.JobPermission target) {
      if ((source == null) || (target == null)) return;
      target.JobId = source.JobId;
      target.GrantedUserId = source.GrantedUserId;
      target.GrantedByUserId = source.GrantedByUserId;
      target.Permission = source.Permission.ToEntity();
    }
    #endregion

    #region State
    public static DT.TaskState ToDto(this DA.TaskState source) {
      switch (source) {
        case DA.TaskState.Aborted: return DT.TaskState.Aborted;
        case DA.TaskState.Calculating: return DT.TaskState.Calculating;
        case DA.TaskState.Failed: return DT.TaskState.Failed;
        case DA.TaskState.Finished: return DT.TaskState.Finished;
        case DA.TaskState.Offline: return DT.TaskState.Offline;
        case DA.TaskState.Paused: return DT.TaskState.Paused;
        case DA.TaskState.Transferring: return DT.TaskState.Transferring;
        case DA.TaskState.Waiting: return DT.TaskState.Waiting;
        default: return DT.TaskState.Failed;
      }
    }

    public static DA.TaskState ToEntity(this DT.TaskState source) {
      switch (source) {
        case DT.TaskState.Aborted: return DA.TaskState.Aborted;
        case DT.TaskState.Calculating: return DA.TaskState.Calculating;
        case DT.TaskState.Failed: return DA.TaskState.Failed;
        case DT.TaskState.Finished: return DA.TaskState.Finished;
        case DT.TaskState.Offline: return DA.TaskState.Offline;
        case DT.TaskState.Paused: return DA.TaskState.Paused;
        case DT.TaskState.Transferring: return DA.TaskState.Transferring;
        case DT.TaskState.Waiting: return DA.TaskState.Waiting;
        default: return DA.TaskState.Failed;
      }
    }
    #endregion

    #region StateLogs
    public static DT.StateLog ToDto(this DA.StateLog source) {
      return new DT.StateLog {
        Id = source.StateLogId,
        State = source.State.ToDto(),
        DateTime = source.DateTime,
        TaskId = source.TaskId,
        UserId = source.UserId,
        SlaveId = source.SlaveId,
        Exception = source.Exception
      };
    }

    public static DA.StateLog ToEntity(this DT.StateLog source) {
      return new DA.StateLog {
        StateLogId = source.Id,
        State = source.State.ToEntity(),
        DateTime = source.DateTime,
        TaskId = source.TaskId,
        UserId = source.UserId,
        SlaveId = source.SlaveId,
        Exception = source.Exception
      };
    }
    #endregion

    #region Command
    public static DT.Command? ToDto(this DA.Command? source) {
      if (source.HasValue) {
        switch (source) {
          case DA.Command.Abort: return DT.Command.Abort;
          case DA.Command.Pause: return DT.Command.Pause;
          case DA.Command.Stop: return DT.Command.Stop;
          default: return DT.Command.Pause;
        }
      }
      return null;
    }

    public static DA.Command? ToEntity(this DT.Command? source) {
      if (source.HasValue) {
        switch (source) {
          case DT.Command.Abort: return DA.Command.Abort;
          case DT.Command.Pause: return DA.Command.Pause;
          case DT.Command.Stop: return DA.Command.Stop;
          default: return DA.Command.Pause;
        }
      }
      return null;
    }
    #endregion

    #region Permission
    public static DT.Permission ToDto(this DA.Permission source) {
      switch (source) {
        case DA.Permission.Full: return DT.Permission.Full;
        case DA.Permission.NotAllowed: return DT.Permission.NotAllowed;
        case DA.Permission.Read: return DT.Permission.Read;
        default: return DT.Permission.NotAllowed;
      }
    }

    public static DA.Permission ToEntity(this DT.Permission source) {
      switch (source) {
        case DT.Permission.Full: return DA.Permission.Full;
        case DT.Permission.NotAllowed: return DA.Permission.NotAllowed;
        case DT.Permission.Read: return DA.Permission.Read;
        default: return DA.Permission.NotAllowed;
      }
    }
    #endregion

  }
}
